#include <stdio.h>
#include <stdlib.h>
#include <avr\io.h>
#include <avr\interrupt.h>
#include <string.h>
#include <util/delay.h>


/**************************************************
 * DEFINE-uri pentru parametrizarea modulului LCD *
 **************************************************/

#define LcdDATA_DDR		DDRA			// Portul pe care conectam firele de date la LCD-ul
#define LcdDATA_PORT	PORTA
#define LcdDATA_PIN		PINA

#define LcdCMD_DDR		DDRA			// Portul pe care conectam firele de comenzi la LCD
#define LcdCMD_PORT		PORTA
#define LcdCMD_PIN		PINA

#define LcdD4			PA0				// Pin-ul pentru firul de date D4 de pe LCD
#define LcdD5			PA1				// Pin-ul pentru firul de date D5 de pe LCD
#define LcdD6			PA2				// Pin-ul pentru firul de date D6 de pe LCD
#define LcdD7			PA3				// Pin-ul pentru firul de date D7 de pe LCD

#define LcdRS			PA4				// Pinul pentru selectare operatie (LCD)
#define LcdRW			PA5				// Pinul pentru Read/ Write (LCD)
#define LcdE			PA6				// Pinul de Enable (LCD)

#define LCD_INSTR_4wire 		0x28 	// 4 fire date, font 5x8
#define LCD_INSTR_display 		0x0C 	// Display On, Cursor On, Blinking On ( 1 Display Cursor Blink )
#define LCD_INSTR_clearDisplay 	0x01 	// Clear Display
#define LCD_INSTR_returnHome 	0x02 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_nextLine 		0xC0 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_gotoCGRAM		0x40	// Go to Character Generator RAM
#define LCD_SET_DDRAM_ADDR      0x80	// Set DDRAM Address

/************************************************
 * DEFINE-uri pentru parametrizarea keypad-ului *
 ************************************************/

#define keypad_PORT 	PORTC           // Keypad Port
#define keypad_DDR		DDRC        	// Data Direction Register
#define keypad_PIN		PINC         	// Keypad Port Pins

/*********************************************************
 * Alte DEFINE-uri pentru variabile utilizate in program *
 *********************************************************/

#define TRUE	1
#define FALSE	0

/***************************************************************
 * API LCD... implementarea este in partea de jos a fisierului *
 ***************************************************************/

void LCD_init();											// Initializare modul LCD.. trebuie apelata inainte de a se face orice operatie cu LCD-ul
void LCD_writeInstruction(unsigned char _instruction);		// Trimite o instructiune catre lcd (vezi datasheet)
void LCD_writeData(unsigned char _data);					// Trimite date catre LCD pentru afisare
void LCD_write(unsigned char _byte);						// trimite un byte catre LCD la modul general (nu conteaza daca e instructiune sau date)
void LCD_waitNotBusy();										// Asteptam pana cand lcd-ul devine disponibil pt o noua comanda
void LCD_waitInstructions(unsigned char _instructions);		// Asteapta un numar de cicli de tact.. loop
void LCD_print(char* _msg);									// Afiseaza imformatia pe LCD (doar 1 linie, primele 16 caractere)
void LCD_gotoXY(unsigned char x, unsigned char y);			// Trimite curosul LCD-ului la pozitia (x, y)

/******************************************************************
 * API keypad... implementarea este in partea de jos a fisierului *
 ******************************************************************/

void key_init();
unsigned char get_key();

/************************************
 * Definirea variabilelor utilizate *
 ************************************/

unsigned char key_val;							// Tasta apasata
unsigned char LineAdress[2] = {0x00, 0x40};		// Adresa celor doua linii ale LCD-ului

int i;
int key_pressed;								// Se verifica daca o tasta a fost apasata

/*********************
 * Codul propriu-zis *
 *********************/

 float adunare(float a, float b) 
{
	return a+b;
}

float scadere(float a, float b) 
{
	return a-b;
}

float inmultire(float a, float b) 
{
	return a*b;
}

float impartire(float a, float b) 
{
	return a/b;
}


int main()
{
	LCD_init();

	key_pressed = FALSE;
	key_val = 0;
	key_init();

	// Initializez caracterele custom
	LCD_writeInstruction(LCD_INSTR_gotoCGRAM);
	LCD_writeInstruction(LCD_INSTR_clearDisplay);
	LCD_gotoXY(0, 0);
	LCD_print("Calculator");
	LCD_gotoXY(0,1);

//	sei();
	
	char a[20]=" ",b[20];
	int dim_a=0;
	unsigned char op=' ';
	float rez=0;

	// Se verifica in permanenta tastele apasate
	while (1)
    {       
		
		key_val = get_key();
        
			if (key_pressed)
			{
				if (key_val == '*') 
				{
					LCD_gotoXY(0,0); 
 					for(i=0; i<16; i++) LCD_writeData(' ');
					LCD_gotoXY(0,1); 
					for(i=0; i<16; i++) LCD_writeData(' ');
					LCD_gotoXY(0,1); 
					op=' ';
					dim_a=0;
					a[0]='\0';
					b[0]='\0';
				}
				else if ((key_val == '/') || (key_val == 'x') || (key_val == '-') || (key_val == '+'))
				{ 
					if (op==' ')
					{
						LCD_gotoXY(0,0); 
						for(i=0; i<16; i++) LCD_writeData(' ');
						LCD_gotoXY(0,0);
						b[0]='\0';
						strcat(b,a);
						b[dim_a]='\0';
						for(i=0; i<dim_a; i++)
						{
							LCD_writeData(a[i]);
						}
 						dim_a=0;
						a[0]='\0';
						LCD_writeData(key_val);
						LCD_gotoXY(0,1); 
						for(i=0; i<16; i++) LCD_writeData(' ');
						LCD_gotoXY(0,1); 
						op=key_val;			
					}
				}
				else if (key_val == '=')
				{
					LCD_gotoXY(0,0); 
					for(i=0; i<16; i++) LCD_writeData(' ');
					LCD_gotoXY(0,1); 
					for(i=0; i<16; i++) LCD_writeData(' ');
					LCD_gotoXY(0,1);
					a[dim_a]='\0';
					if (op=='/')
					{
						rez=impartire(atof(b),atof(a));
					}
					if (op=='x')
					{
						rez=inmultire(atof(b),atof(a));
					}
					if (op=='-')
					{
						rez=scadere(atof(b),atof(a));
					}
					if (op=='+')
					{
						rez=adunare(atof(b),atof(a));
					}

					long x=(long) rez;
					char aux[20]="";
					int dim_aux=0;
					long rap=x/10;
					while (rap!=0)
					{
						aux[dim_aux]='0'+(x%10);
						x=x/10;
						rap=x/10;
						dim_aux++;
						
					}
					aux[dim_aux]='0'+(x%10);
					dim_aux++;
					for (i=dim_aux-1;i>=0;i--)
					{
						a[dim_aux-i-1]=aux[i];
					}
					dim_a=dim_aux;
					int y=(int) (rez-x)*1000000;
					if (y!=0) {
						a[dim_a]='.';
						dim_a++;
						dim_aux=0;
						rap=y/10;
						while (rap!=0)
						{
							aux[dim_aux]='0'+(y%10);
							y=y/10;
							rap=y/10;
							dim_aux++;
						
						}
						aux[dim_aux]='0'+(y%10);
						for (i=dim_aux-1;i>=0;i--)
						{
							a[dim_a+dim_aux-i-1]=aux[i];
						}	
						dim_a=dim_a+dim_aux;
					}
				
					for(i=0; i<dim_a; i++)
					{
						LCD_writeData(a[i]);
					}
					op=' ';
				}
				else {
					LCD_writeData(key_val);
					a[dim_a]=key_val;
					dim_a++;
				}

			}
		key_pressed = FALSE;
    }
	return 0;
} 

/************************ 
 * Implementare API LCD *
 ************************/

void LCD_init()
{
	LcdDATA_DDR |=  (1<<LcdD4)|(1<<LcdD5)|(1<<LcdD6)|(1<<LcdD7);	// setam pinii de date ca pini de iesire
	LcdCMD_DDR  |=  (1<<LcdRS)|(1<<LcdRW)|(1<<LcdE);				// setam pinii de comenzi ca pini de iesire

	LCD_waitNotBusy();
	
	LcdCMD_PORT   &= ~(1<<LcdRS);								// Setam linia RS pe low
	LcdCMD_PORT   &= ~(1<<LcdRW);								// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD4)&~(1<<LcdD6)&~(1<<LcdD7); 		// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |=  (1<<LcdD5);								// setam pinii de comenzi ca pini de iesire
	
	LcdCMD_PORT |=  (1<<LcdE);					// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	LCD_waitInstructions(6);					// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE );					// Setam linia E(nable) pe low; transferul s-a terminat

	LCD_writeInstruction(LCD_INSTR_4wire); 		// Incarcam comanda: 4 bit data, 2 lines, 5x8 font
	LCD_writeInstruction(LCD_INSTR_display); 	// Display On, Cursor On, Blinking On
	LCD_writeInstruction(0x06);					// Increment, no shift
	LCD_writeInstruction(0x01);					// Clear Display
}

void LCD_writeInstruction(unsigned char _instruction)
{
	LCD_waitNotBusy();						// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	LcdCMD_PORT &= ~(1 << LcdRS);			// setam pinul RS pe low.. low=instructiuni, high=date
	LcdCMD_PORT &= ~(1 << LcdRW);			// setam pinul RW pe low (suntem in modul de comenzi acum)
	LCD_write(_instruction);				// apelam procedura ce trimite byte-ul pe firele de date
}


void LCD_writeData(unsigned char _data)
{
	LCD_waitNotBusy();						// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	LcdCMD_PORT |= (1 << LcdRS);			// setam pinul RS pe high
	LcdCMD_PORT &= ~(1 << LcdRW);			// setam pinul RW pe low (suntem in modul de comenzi acum)
	LCD_write(_data);						// apelam procedura ce trimite byte-ul pe firele de date
}

void LCD_write(unsigned char _byte)
{
	unsigned char _byte2;
	
	_byte2 = _byte>>4;

	LcdDATA_PORT &= ~(1<<LcdD4);
	if ( bit_is_set( _byte2, 0 ) )
		LcdDATA_PORT |= (1<<LcdD4);
		
	LcdDATA_PORT &= ~(1<<LcdD5);
	if ( bit_is_set( _byte2, 1 ) )
		LcdDATA_PORT |= (1<<LcdD5);

	LcdDATA_PORT &= ~(1<<LcdD6);
	if ( bit_is_set( _byte2, 2 ) )
		LcdDATA_PORT |= (1<<LcdD6);

	LcdDATA_PORT &= ~(1<<LcdD7);
	if ( bit_is_set( _byte2, 3 ) )
		LcdDATA_PORT |= (1<<LcdD7);

	LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
	LCD_waitInstructions(6);						// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

	LCD_waitInstructions(6);						// Asteptam o perioada de timp T
	
	LcdDATA_PORT &= ~(1<<LcdD4);
	if ( bit_is_set( _byte, 0 ) )
		LcdDATA_PORT |= (1<<LcdD4);
	
	LcdDATA_PORT &= ~(1<<LcdD5);
	if ( bit_is_set( _byte, 1 ) )
		LcdDATA_PORT |= (1<<LcdD5);

	LcdDATA_PORT &= ~(1<<LcdD6);
	if ( bit_is_set( _byte, 2 ) )
		LcdDATA_PORT |= (1<<LcdD6);

	LcdDATA_PORT &= ~(1<<LcdD7);
	if ( bit_is_set( _byte, 3 ) )
		LcdDATA_PORT |= (1<<LcdD7);

	LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
	LCD_waitInstructions(6);						// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low
}

void LCD_waitNotBusy()
{
	unsigned char _loop = 1;

	while (_loop)
	{
		LcdDATA_DDR &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7);	// Setam pinii de date de la LCD pe in pt a citi busy flag
		LcdDATA_PORT &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); 	// Dezactivam pullup resistor pentru pinii de in
		
		LcdCMD_PORT &= ~(1<<LcdE);					// Setam pin-ul e pe low; ar trebui sa fie deja low, doar ne asiguram
		LcdCMD_PORT &= ~(1<<LcdRS);					// Setam pinul RS pe low
		LcdCMD_PORT |=  (1<<LcdRW);					// Setam pinul RW pe high (acum suntem in modul de interogare busy/adr)
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		_loop = LcdDATA_PIN & (1<<LcdD7);				// Citim busy flag-ul
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low
		
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

		LcdDATA_DDR |= (1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); // Setam Portul de LCD ca port de iesire la loc
	}
}

void LCD_waitInstructions(unsigned char _instructions)
{
	while (_instructions--)
		;
}

void LCD_print(char* _msg)
{
	unsigned char i=0;
	for( ; _msg[i]!=0 && i<16; i++)
		LCD_writeData( _msg[i] );
}

void LCD_gotoXY(unsigned char x, unsigned char y)
{
    LCD_writeInstruction(LCD_SET_DDRAM_ADDR | (LineAdress[y] + x));		// Se trimite instructiunea necesara avansarii cursorului la pozitia indicata
}

/*************************** 
 * Implementare API keypad *
 ***************************/


void key_init()
{
	keypad_DDR = 0x0F;			
	keypad_PORT = 0xF0;		
}

unsigned char get_key(){
	unsigned char key='0';

	_delay_ms(10);
	key_init();
	// Prima coloana
	keypad_PORT = 0b11111110;
	if (!bit_is_set(keypad_PIN, 4)) {while(!bit_is_set(keypad_PIN, 4)); key_pressed = TRUE; key = '1';}	// Linia 1
	if (!bit_is_set(keypad_PIN, 5)) {while(!bit_is_set(keypad_PIN, 5)); key_pressed = TRUE; key = '4';}	// Linia 2
	if (!bit_is_set(keypad_PIN, 6)) {while(!bit_is_set(keypad_PIN, 6)); key_pressed = TRUE; key = '7';}	// Linia 3
	if (!bit_is_set(keypad_PIN, 7)) {while(!bit_is_set(keypad_PIN, 7)); key_pressed = TRUE; key = '*';}	// Linia 4
  
	key_init();
	// A doua coloana
	keypad_PORT = 0b11111101;
	if (!bit_is_set(keypad_PIN, 4)) {while(!bit_is_set(keypad_PIN, 4)); key_pressed = TRUE; key = '2';}
	if (!bit_is_set(keypad_PIN, 5)) {while(!bit_is_set(keypad_PIN, 5)); key_pressed = TRUE; key = '5';}
	if (!bit_is_set(keypad_PIN, 6)) {while(!bit_is_set(keypad_PIN, 6)); key_pressed = TRUE; key = '8';}
	if (!bit_is_set(keypad_PIN, 7)) {while(!bit_is_set(keypad_PIN, 7)); key_pressed = TRUE; key = '0';}
  
	key_init();
	// A treia coloana
	keypad_PORT = 0b11111011;
	if (!bit_is_set(keypad_PIN, 4)) {while(!bit_is_set(keypad_PIN, 4)); key_pressed = TRUE; key = '3';}
	if (!bit_is_set(keypad_PIN, 5)) {while(!bit_is_set(keypad_PIN, 5)); key_pressed = TRUE; key = '6';}
	if (!bit_is_set(keypad_PIN, 6)) {while(!bit_is_set(keypad_PIN, 6)); key_pressed = TRUE; key = '9';}
	if (!bit_is_set(keypad_PIN, 7)) {while(!bit_is_set(keypad_PIN, 7)); key_pressed = TRUE; key = '=';}
 
	key_init();
	// A patra coloana
	keypad_PORT = 0b11110111;
	if (!bit_is_set(keypad_PIN, 4)) {while(!bit_is_set(keypad_PIN, 4)); key_pressed = TRUE; key = '/';} 
	if (!bit_is_set(keypad_PIN, 5)) {while(!bit_is_set(keypad_PIN, 5)); key_pressed = TRUE; key = 'x';} 
	if (!bit_is_set(keypad_PIN, 6)) {while(!bit_is_set(keypad_PIN, 6)); key_pressed = TRUE; key = '-';}
	if (!bit_is_set(keypad_PIN, 7)) {while(!bit_is_set(keypad_PIN, 7)); key_pressed = TRUE; key = '+';}
  
	key_init();

	return key;
}





